from __future__ import annotations

import os
from contextvars import ContextVar
from pathlib import Path
from textwrap import dedent
from typing import Optional

import attr

from .code_file import CodeFile, CodeMessageLevel, count_feature_tokens
from .code_file_manager import CODE_FILE_MANAGER
from .code_map import check_ctags_disabled
from .diff_context import DiffContext
from .embeddings import get_feature_similarity_scores
from .git_handler import GIT_ROOT, get_non_gitignored_files, get_paths_with_git_diffs
from .include_files import (
    build_path_tree,
    get_include_files,
    is_file_text_encoded,
    print_path_tree,
)
from .llm_api import count_tokens
from .session_stream import SESSION_STREAM
from .utils import sha256


@attr.define
class CodeContextSettings:
    diff: Optional[str] = None
    pr_diff: Optional[str] = None
    no_code_map: bool = False
    use_embedding: bool = False
    auto_tokens: Optional[int] = None


CODE_CONTEXT: ContextVar[CodeContext] = ContextVar("mentat:code_context")


class CodeContext:
    settings: CodeContextSettings
    include_files: dict[Path, CodeFile]
    diff_context: DiffContext
    code_map: bool = True
    features: list[CodeFile] = []

    def __init__(
        self,
        settings: CodeContextSettings,
    ):
        self.settings = settings

    @classmethod
    async def create(
        cls, paths: list[Path], exclude_paths: list[Path], settings: CodeContextSettings
    ):
        stream = SESSION_STREAM.get()

        self = cls(settings)

        self.diff_context = await DiffContext.create(
            self.settings.diff, self.settings.pr_diff
        )
        self.include_files, invalid_paths = get_include_files(paths, exclude_paths)
        for invalid_path in invalid_paths:
            await stream.send(
                f"File path {invalid_path} is not text encoded, and was skipped.",
                color="light_yellow",
            )
        await self._set_code_map()

        return self

    async def _set_code_map(self):
        stream = SESSION_STREAM.get()

        if self.settings.no_code_map:
            self.code_map = False
        else:
            disabled_reason = await check_ctags_disabled()
            if disabled_reason:
                ctags_disabled_message = f"""
                    There was an error with your universal ctags installation, disabling CodeMap.
                    Reason: {disabled_reason}
                """
                ctags_disabled_message = dedent(ctags_disabled_message)
                await stream.send(ctags_disabled_message, color="yellow")
                self.settings.no_code_map = True
                self.code_map = False
            else:
                self.code_map = True

    async def display_context(self):
        """Display the baseline context: included files and auto-context settings"""
        stream = SESSION_STREAM.get()
        git_root = GIT_ROOT.get()

        await stream.send("\nCode Context:", color="blue")
        prefix = "  "
        await stream.send(f"{prefix}Directory: {git_root}")
        if self.diff_context.name:
            await stream.send(f"{prefix}Diff:", end=" ")
            await stream.send(self.diff_context.get_display_context(), color="green")
        if self.include_files:
            await stream.send(f"{prefix}Included files:")
            await stream.send(f"{prefix + prefix}{git_root.name}")
            await print_path_tree(
                build_path_tree(list(self.include_files.values()), git_root),
                get_paths_with_git_diffs(),
                git_root,
                prefix + prefix,
            )
        else:
            await stream.send(f"{prefix}Included files: None", color="yellow")
        await stream.send(
            f"{prefix}CodeMaps: {'Enabled' if self.code_map else 'Disabled'}"
        )
        auto = self.settings.auto_tokens
        await stream.send(
            f"{prefix}Auto-tokens: {'Model max (default)' if auto is None else auto}"
        )

    async def display_features(self):
        """Display a summary of all active features"""
        auto_features = {level: 0 for level in CodeMessageLevel}
        for f in self.features:
            if f.path not in self.include_files:
                auto_features[f.level] += 1
        if any(auto_features.values()):
            stream = SESSION_STREAM.get()
            await stream.send("Auto-Selected Features:", color="blue")
            for level, count in auto_features.items():
                if count:
                    await stream.send(f"  {count} {level.description}")

    _code_message: str | None = None
    _code_message_checksum: str | None = None

    def _get_code_message_checksum(self, max_tokens: Optional[int] = None) -> str:
        if not self.features:
            features_checksum = ""
        else:
            git_root = GIT_ROOT.get()
            code_file_manager = CODE_FILE_MANAGER.get()
            feature_files = {Path(git_root / f.path) for f in self.features}
            feature_file_checksums = [
                code_file_manager.get_file_checksum(f) for f in feature_files
            ]
            features_checksum = sha256("".join(feature_file_checksums))
        settings = attr.asdict(self.settings)
        settings["max_tokens"] = max_tokens
        settings["include_files"] = self.include_files
        settings_checksum = sha256(str(settings))
        return features_checksum + settings_checksum

    async def get_code_message(
        self,
        prompt: str,
        model: str,
        max_tokens: int,
    ) -> str:
        code_message_checksum = self._get_code_message_checksum(max_tokens)
        if (
            self._code_message is None
            or code_message_checksum != self._code_message_checksum
        ):
            self._code_message = await self._get_code_message(prompt, model, max_tokens)
            self._code_message_checksum = self._get_code_message_checksum(max_tokens)
        return self._code_message

    async def _get_code_message(
        self,
        prompt: str,
        model: str,
        max_tokens: int,
    ) -> str:
        code_message = list[str]()

        self.diff_context.clear_cache
        await self._set_code_map()
        if self.diff_context.files:
            code_message += [
                "Diff References:",
                f' "-" = {self.diff_context.name}',
                ' "+" = Active Changes',
                "",
            ]
        code_message += ["Code Files:\n"]

        features = self._get_include_features()
        include_feature_tokens = sum(await count_feature_tokens(features, model))
        include_feature_tokens -= count_tokens("\n".join(code_message), model)
        _max_auto = max(0, max_tokens - include_feature_tokens)
        _max_user = self.settings.auto_tokens
        if _max_auto == 0 or _max_user == 0:
            self.features = features
        else:
            auto_tokens = _max_auto if _max_user is None else min(_max_auto, _max_user)
            self.features = await self._get_auto_features(
                prompt, model, features, auto_tokens
            )

        for f in self.features:
            code_message += await f.get_code_message()
        return "\n".join(code_message)

    def _get_include_features(self) -> list[CodeFile]:
        git_root = GIT_ROOT.get()
        include_features = list[CodeFile]()
        for path, feature in self.include_files.items():
            if feature.level == CodeMessageLevel.INTERVAL:
                interval_str = ",".join(f"{i.start}-{i.end}" for i in feature.intervals)
                path = f"{path}:{interval_str}"
            diff_target = (
                self.diff_context.target if path in self.diff_context.files else None
            )
            feature = CodeFile(path, feature.level, diff=diff_target)
            include_features.append(feature)

        def _feature_relative_path(f: CodeFile) -> str:
            return os.path.relpath(f.path, git_root)

        return sorted(include_features, key=_feature_relative_path)

    async def _get_auto_features(
        self,
        prompt: str,
        model: str,
        include_features: list[CodeFile],
        max_tokens: int,
    ) -> list[CodeFile]:
        git_root = GIT_ROOT.get()

        # Find the first (longest) level that fits
        include_features_tokens = sum(
            await count_feature_tokens(include_features, model)
        )
        max_auto_tokens = max_tokens - include_features_tokens
        all_features = include_features.copy()
        levels = [CodeMessageLevel.FILE_NAME]
        if not self.settings.no_code_map:
            levels = [CodeMessageLevel.CMAP_FULL, CodeMessageLevel.CMAP] + levels
        for level in levels:
            _features = list[CodeFile]()
            for path in get_non_gitignored_files(git_root):
                if (
                    path in self.include_files
                    or path.is_dir()
                    or not is_file_text_encoded(path)
                ):
                    continue
                diff_target = (
                    self.diff_context.target
                    if path in self.diff_context.files
                    else None
                )
                feature = CodeFile(path, level=level, diff=diff_target)
                _features.append(feature)
            level_length = sum(await count_feature_tokens(_features, model))
            if level_length < max_auto_tokens:
                all_features += _features
                break

        # Sort by relative path
        def _feature_relative_path(f: CodeFile) -> str:
            return os.path.relpath(f.path, git_root)

        all_features = sorted(all_features, key=_feature_relative_path)

        # If there's room, convert cmap features to code features (full text)
        # starting with the highest-scoring.
        cmap_features_tokens = sum(await count_feature_tokens(all_features, model))
        max_sim_tokens = max_tokens - cmap_features_tokens
        if self.settings.auto_tokens is not None:
            max_sim_tokens = min(max_sim_tokens, self.settings.auto_tokens)

        if self.settings.use_embedding and max_sim_tokens > 0 and prompt != "":
            sim_tokens = 0

            # Get embedding-similarity scores for all files
            all_code_features = [
                CodeFile(f.path, CodeMessageLevel.CODE, f.diff)
                for f in all_features
                if f.path not in self.include_files
            ]
            sim_scores = await get_feature_similarity_scores(prompt, all_code_features)
            all_code_features_scored = zip(all_code_features, sim_scores)
            all_code_features_sorted = sorted(
                all_code_features_scored, key=lambda x: x[1], reverse=True
            )
            for code_feature, _ in all_code_features_sorted:
                # Calculate the total change in length
                i_cmap, cmap_feature = next(
                    (i, f)
                    for i, f in enumerate(all_features)
                    if f.path == code_feature.path
                )
                recovered_tokens = await cmap_feature.count_tokens(model)
                new_tokens = await code_feature.count_tokens(model)
                forecast = max_sim_tokens - sim_tokens + recovered_tokens - new_tokens
                if forecast > 0:
                    sim_tokens = sim_tokens + new_tokens - recovered_tokens
                    all_features[i_cmap] = code_feature

        return sorted(all_features, key=_feature_relative_path)

    def include_file(self, path: Path):
        paths, invalid_paths = get_include_files([path], [])
        for new_path, new_file in paths.items():
            if new_path not in self.include_files:
                self.include_files[new_path] = new_file
        return invalid_paths

    def exclude_file(self, path: Path):
        paths, _ = get_include_files([path], [])
        for new_path in paths.keys():
            if new_path in self.include_files:
                del self.include_files[new_path]



from __future__ import annotations

import json
import logging
from contextvars import ContextVar
from enum import Enum
from timeit import default_timer

from openai.error import InvalidRequestError, RateLimitError

from mentat.parsers.file_edit import FileEdit
from mentat.parsers.parser import PARSER, Parser
from tests.conftest import SessionStream

from .code_context import CODE_CONTEXT
from .config_manager import CONFIG_MANAGER, user_config_path
from .errors import MentatError, UserError
from .llm_api import (
    COST_TRACKER,
    call_llm_api,
    count_tokens,
    get_prompt_token_count,
    is_model_available,
    model_context_size,
)
from .session_stream import SESSION_STREAM

CONVERSATION: ContextVar[Conversation] = ContextVar("mentat:conversation")


class MessageRole(Enum):
    System = "system"
    User = "user"
    Assistant = "assistant"


class Conversation:
    max_tokens: int

    def __init__(self):
        config = CONFIG_MANAGER.get()
        parser = PARSER.get()

        self.model = config.model()
        self.messages = list[dict[str, str]]()

        # This contain the messages the user actually sends and the messages the model output
        # along with a snapshot of exactly what the model got before that message
        self.literal_messages = list[tuple[str, list[dict[str, str]] | None]]()

        prompt = parser.get_system_prompt()
        self.add_message(MessageRole.System, prompt)

    async def display_token_count(self):
        stream = SESSION_STREAM.get()
        parser = PARSER.get()
        code_context = CODE_CONTEXT.get()
        config = CONFIG_MANAGER.get()

        if not is_model_available(self.model):
            raise MentatError(
                f"Model {self.model} is not available. Please try again with a"
                " different model."
            )
        if "gpt-4" not in self.model:
            await stream.send(
                "Warning: Mentat has only been tested on GPT-4. You may experience"
                " issues with quality. This model may not be able to respond in"
                " mentat's edit format.",
                color="yellow",
            )
            if "gpt-3.5" not in self.model:
                await stream.send(
                    "Warning: Mentat does not know how to calculate costs or context"
                    " size for this model.",
                    color="yellow",
                )
        prompt = parser.get_system_prompt()
        context_size = model_context_size(self.model)
        maximum_context = config.maximum_context()
        if maximum_context:
            if context_size:
                context_size = min(context_size, maximum_context)
            else:
                context_size = maximum_context
        tokens = count_tokens(
            await code_context.get_code_message("", self.model, max_tokens=0),
            self.model,
        ) + count_tokens(prompt, self.model)

        if not context_size:
            raise MentatError(
                f"Context size for {self.model} is not known. Please set"
                f" maximum-context in {user_config_path}."
            )
        else:
            self.max_tokens = context_size
        if context_size and tokens > context_size:
            raise MentatError(
                f"Included files already exceed token limit ({tokens} /"
                f" {context_size}). Please try running again with a reduced"
                " number of files."
            )
        elif tokens + 1000 > context_size:
            await stream.send(
                f"Warning: Included files are close to token limit ({tokens} /"
                f" {context_size}), you may not be able to have a long"
                " conversation.",
                color="red",
            )
        else:
            await stream.send(
                f"Prompt and included files token count: {tokens} / {context_size}",
                color="cyan",
            )

    # The transcript logger logs tuples containing the actual message sent by the user or LLM
    # and (for LLM messages) the LLM conversation that led to that LLM response
    def add_user_message(self, message: str):
        """Used for actual user input messages"""
        transcript_logger = logging.getLogger("transcript")
        transcript_logger.info(json.dumps((message, None)))
        self.literal_messages.append((message, None))
        self.add_message(MessageRole.User, message)

    def add_model_message(self, message: str, messages_snapshot: list[dict[str, str]]):
        """Used for actual model output messages"""
        transcript_logger = logging.getLogger("transcript")
        transcript_logger.info(json.dumps((message, messages_snapshot)))
        self.literal_messages.append((message, messages_snapshot))
        self.add_message(MessageRole.Assistant, message)

    def add_message(self, role: MessageRole, message: str):
        """Used for adding messages to the models conversation"""
        self.messages.append({"role": role.value, "content": message})

    async def _stream_model_response(
        self,
        stream: SessionStream,
        parser: Parser,
        messages: list[dict[str, str]],
    ):
        start_time = default_timer()
        try:
            response = await call_llm_api(messages, self.model)
            await stream.send(
                "Streaming... use control-c to interrupt the model at any point\n"
            )
            async with parser.interrupt_catcher():
                parsedLLMResponse = await parser.stream_and_parse_llm_response(response)
        except InvalidRequestError as e:
            raise MentatError(
                "Something went wrong - invalid request to OpenAI API. OpenAI"
                " returned:\n"
                + str(e)
            )
        except RateLimitError as e:
            raise UserError("OpenAI gave a rate limit error:\n" + str(e))

        time_elapsed = default_timer() - start_time
        return (parsedLLMResponse, time_elapsed)

    async def get_model_response(self) -> list[FileEdit]:
        stream = SESSION_STREAM.get()
        code_context = CODE_CONTEXT.get()
        cost_tracker = COST_TRACKER.get()
        parser = PARSER.get()

        messages_snapshot = self.messages.copy()

        # Rebuild code context with active code and available tokens
        conversation_history = "\n".join([m["content"] for m in messages_snapshot])
        tokens = count_tokens(conversation_history, self.model)
        response_buffer = 1000
        code_message = await code_context.get_code_message(
            messages_snapshot[-1]["content"],
            self.model,
            self.max_tokens - tokens - response_buffer,
        )
        messages_snapshot.append({"role": "system", "content": code_message})

        await code_context.display_features()
        num_prompt_tokens = await get_prompt_token_count(messages_snapshot, self.model)
        parsedLLMResponse, time_elapsed = await self._stream_model_response(
            stream, parser, messages_snapshot
        )
        await cost_tracker.display_api_call_stats(
            num_prompt_tokens,
            count_tokens(parsedLLMResponse.full_response, self.model),
            self.model,
            time_elapsed,
        )

        messages_snapshot.append(
            {"role": "assistant", "content": parsedLLMResponse.full_response}
        )
        self.add_model_message(parsedLLMResponse.full_response, messages_snapshot)
        return parsedLLMResponse.file_edits



import os
from pathlib import Path

import attr
from termcolor import colored


# All paths should be abs paths
@attr.define()
class RenameAction:
    old_file_name: Path = attr.field()
    cur_file_name: Path = attr.field()

    def undo(self) -> str | None:
        if self.old_file_name.exists():
            return colored(
                f"File {self.old_file_name} already exists; unable to undo rename from"
                f" {self.cur_file_name}",
                color="light_red",
            )
        else:
            os.rename(self.cur_file_name, self.old_file_name)


@attr.define()
class CreationAction:
    cur_file_name: Path = attr.field()

    def undo(self) -> str | None:
        if not self.cur_file_name.exists():
            return colored(
                f"File {self.cur_file_name} does not exist; unable to delete",
                color="light_red",
            )
        else:
            self.cur_file_name.unlink()


@attr.define()
class DeletionAction:
    old_file_name: Path = attr.field()
    old_file_lines: list[str] = attr.field()

    def undo(self) -> str | None:
        if self.old_file_name.exists():
            return colored(
                f"File {self.old_file_name} already exists; unable to re-create",
                color="light_red",
            )
        else:
            with open(self.old_file_name, "w") as f:
                f.write("\n".join(self.old_file_lines))


@attr.define()
class EditAction:
    cur_file_name: Path = attr.field()
    old_file_lines: list[str] = attr.field()

    def undo(self) -> str | None:
        if not self.cur_file_name.exists():
            return colored(
                f"File {self.cur_file_name} does not exist; unable to undo edit",
                color="light_red",
            )
        else:
            with open(self.cur_file_name, "w") as f:
                f.write("\n".join(self.old_file_lines))


HistoryAction = RenameAction | CreationAction | DeletionAction | EditAction


# TODO: Keep track of when we create directories so we can undo those as well
class EditHistory:
    def __init__(self):
        self.edits = list[list[HistoryAction]]()
        self.cur_edit = list[HistoryAction]()

    def add_action(self, history_action: HistoryAction):
        self.cur_edit.append(history_action)

    def push_edits(self):
        if self.cur_edit:
            self.edits.append(self.cur_edit)
            self.cur_edit = list[HistoryAction]()

    # TODO: Add redo
    def undo(self) -> str:
        if not self.edits:
            return colored("No edits available to undo", color="light_red")

        # Make sure to go top down
        cur_edit = self.edits.pop()
        errors = list[str]()
        while cur_edit:
            cur_action = cur_edit.pop()
            error = cur_action.undo()
            if error is not None:
                errors.append(error)
        return "\n".join(errors)

    def undo_all(self) -> str:
        if not self.edits:
            return colored("No edits available to undo", color="light_red")

        errors = list[str]()
        while self.edits:
            error = self.undo()
            if error:
                errors.append(error)
        return "\n".join(errors)



import asyncio
import hashlib
import json
from importlib import resources
from importlib.abc import Traversable
from pathlib import Path
from typing import AsyncGenerator

mentat_dir_path = Path.home() / ".mentat"

# package_name should always be "mentat" - but this will work if package name is changed
package_name = __name__.split(".")[0]
resources_path = Path("resources")
conversation_viewer_path = Path("conversation_viewer.html")


def sha256(data: str) -> str:
    return hashlib.sha256(data.encode("utf-8")).hexdigest()


async def run_subprocess_async(*args: str) -> str:
    process = await asyncio.create_subprocess_exec(
        *args,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        start_new_session=True,
    )
    stdout, stderr = await process.communicate()

    if process.returncode != 0:
        error_output = stderr.decode("utf-8").strip() if stderr else ""
        raise Exception(f"Subprocess failed with error: {error_output}")

    output = stdout.decode("utf-8").strip() if stdout else ""

    return output


# Useful for using functions designed to work with LLMs on prepared strings
async def convert_string_to_asyncgen(
    input_str: str, chunk_size: int
) -> AsyncGenerator[dict[str, list[dict[str, dict[str, str]]]], None]:
    for i in range(0, len(input_str), chunk_size):
        yield {"choices": [{"delta": {"content": input_str[i : i + chunk_size]}}]}
    return


def fetch_resource(resource_path: Path) -> Traversable:
    resource = resources.files(package_name).joinpath(
        str(resources_path / resource_path)
    )
    return resource


# TODO: Should we use a templating library (like jinja?) for this?
def create_viewer(
    literal_messages: list[tuple[str, list[dict[str, str]] | None]]
) -> Path:
    messages_json = json.dumps(literal_messages)
    viewer_resource = fetch_resource(conversation_viewer_path)
    with viewer_resource.open("r") as viewer_file:
        html = viewer_file.read()
    html = html.replace("{{ messages }}", messages_json)

    viewer_path = mentat_dir_path / conversation_viewer_path
    with viewer_path.open("w") as viewer_file:
        viewer_file.write(html)
    return viewer_path



import subprocess
from pathlib import Path
from textwrap import dedent

import pytest

from mentat.code_context import CODE_CONTEXT
from mentat.commands import Command, HelpCommand, InvalidCommand
from mentat.session import Session


def test_invalid_command():
    assert isinstance(Command.create_command("non-existent"), InvalidCommand)


@pytest.mark.asyncio
async def test_help_command(mock_stream):
    command = Command.create_command("help")
    await command.apply()
    assert isinstance(command, HelpCommand)


@pytest.mark.asyncio
async def test_commit_command(
    temp_testbed, mock_setup_api_key, mock_collect_user_input
):
    file_name = "test_file.py"
    with open(file_name, "w") as f:
        f.write("# Commit me!")

    mock_collect_user_input.set_stream_messages(
        [
            "/commit",
            "q",
        ]
    )

    session = await Session.create([])
    await session.start()
    await session.stream.stop()

    assert subprocess.check_output(["git", "status", "-s"], text=True) == ""


@pytest.mark.asyncio
async def test_include_command(
    temp_testbed, mock_setup_api_key, mock_collect_user_input
):
    mock_collect_user_input.set_stream_messages(
        [
            "/include scripts",
            "q",
        ]
    )

    session = await Session.create([])
    await session.start()
    await session.stream.stop()

    code_context = CODE_CONTEXT.get()
    assert (
        Path(temp_testbed) / "scripts" / "calculator.py" in code_context.include_files
    )


@pytest.mark.asyncio
async def test_exclude_command(
    temp_testbed, mock_setup_api_key, mock_collect_user_input
):
    mock_collect_user_input.set_stream_messages(
        [
            "/exclude scripts",
            "q",
        ]
    )

    session = await Session.create(["scripts"])
    await session.start()
    await session.stream.stop()

    code_context = CODE_CONTEXT.get()
    print(code_context.include_files)
    assert not code_context.include_files


@pytest.mark.asyncio
async def test_undo_command(
    temp_testbed, mock_setup_api_key, mock_collect_user_input, mock_call_llm_api
):
    temp_file_name = "temp.py"
    with open(temp_file_name, "w") as f:
        f.write(dedent("""\
            # This is a temporary file
            # with 2 lines"""))

    mock_collect_user_input.set_stream_messages(
        [
            "",
            "y",
            "/undo",
            "q",
        ]
    )

    mock_call_llm_api.set_generator_values([dedent(f"""\
        Conversation

        @@start
        {{
            "file": "{temp_file_name}",
            "action": "insert",
            "insert-after-line": 1,
            "insert-before-line": 2
        }}
        @@code
        # I inserted this comment
        @@end""")])

    session = await Session.create([temp_file_name])
    await session.start()
    await session.stream.stop()

    with open(temp_file_name, "r") as f:
        content = f.read()
        expected_content = dedent("""\
            # This is a temporary file
            # with 2 lines""")
    assert content == expected_content


@pytest.mark.asyncio
async def test_undo_all_command(
    temp_testbed, mock_setup_api_key, mock_collect_user_input, mock_call_llm_api
):
    temp_file_name = "temp.py"
    with open(temp_file_name, "w") as f:
        f.write(dedent("""\
            # This is a temporary file
            # with 2 lines"""))

    mock_collect_user_input.set_stream_messages(
        [
            "",
            "y",
            "/undo-all",
            "q",
        ]
    )

    # TODO: Make a way to set multiple return values for call_llm_api and reset multiple edits at once
    mock_call_llm_api.set_generator_values([dedent(f"""\
        Conversation

        @@start
        {{
            "file": "{temp_file_name}",
            "action": "insert",
            "insert-after-line": 1,
            "insert-before-line": 2
        }}
        @@code
        # I inserted this comment
        @@end""")])

    session = await Session.create([temp_file_name])
    await session.start()
    await session.stream.stop()

    with open(temp_file_name, "r") as f:
        content = f.read()
        expected_content = dedent("""\
            # This is a temporary file
            # with 2 lines""")
    assert content == expected_content



import pytest

from mentat.code_file import CodeFile
from mentat.embeddings import _batch_ffd, get_feature_similarity_scores


def test_batch_ffd():
    data = {"a": 4, "b": 5, "c": 3, "d": 2}
    batch_size = 6
    result = _batch_ffd(data, batch_size)
    expected = [["b"], ["a", "d"], ["c"]]
    assert result == expected


def _make_code_file(path, text):
    with open(path, "w") as f:
        f.write(text)
    return CodeFile(path)


@pytest.mark.asyncio
async def test_get_feature_similarity_scores(
    mocker, mock_git_root, mock_code_file_manager, mock_parser, mock_stream
):
    prompt = "example prompt"
    features = [_make_code_file(f"file{i}.txt", f"File {i}") for i in range(3)]
    mocker.patch(
        "mentat.embeddings.call_embedding_api",
        return_value=[
            [0.4, 0.4, 0.4],
            [0.5, 0.6, 0.7],
            [0.69, 0.7, 0.71],
            [0.7, 0.7, 0.7],  # The prompt
        ],
    )
    result = await get_feature_similarity_scores(prompt, features)
    assert len(result) == 3
    assert max(result) == result[0]  # The first feature is most similar



from __future__ import annotations

import json
import logging
from contextvars import ContextVar
from json import JSONDecodeError
from pathlib import Path
from typing import Any, Dict, Optional, cast

from mentat.git_handler import GIT_ROOT
from mentat.session_stream import SESSION_STREAM
from mentat.utils import fetch_resource, mentat_dir_path

default_config_file_name = Path("default_config.json")
config_file_name = Path(".mentat_config.json")
user_config_path = mentat_dir_path / config_file_name

CONFIG_MANAGER: ContextVar[ConfigManager] = ContextVar("mentat:config_manager")


class ConfigManager:
    def __init__(
        self,
        user_config: Dict[str, str],
        project_config: Dict[str, str],
    ):
        self.user_config = user_config
        self.project_config = project_config

        default_config_path = fetch_resource(default_config_file_name)
        with default_config_path.open("r") as config_file:
            self.default_config = json.load(config_file)

    @classmethod
    async def create(cls):
        stream = SESSION_STREAM.get()

        if user_config_path.exists():
            with open(user_config_path) as config_file:
                try:
                    user_config = json.load(config_file)
                except JSONDecodeError:
                    logging.info("User config file contains invalid json")
                    await stream.send(
                        "Warning: User .mentat_config.json contains invalid"
                        " json; ignoring user configuration file",
                        color="light_yellow",
                    )
                    user_config = dict[str, str]()
        else:
            user_config = dict[str, str]()

        project_config_path = GIT_ROOT.get() / config_file_name
        if project_config_path.exists():
            with open(project_config_path) as config_file:
                try:
                    project_config = json.load(config_file)
                except JSONDecodeError:
                    logging.info("Project config file contains invalid json")
                    await stream.send(
                        "Warning: Git project .mentat_config.json contains invalid"
                        " json; ignoring project configuration file",
                        color="light_yellow",
                    )
                    project_config = dict[str, str]()
        else:
            project_config = dict[str, str]()

        self = cls(user_config, project_config)

        return self

    def input_style(self) -> list[tuple[str, str]]:
        return cast(list[tuple[str, str]], self._get_key("input-style"))

    def model(self) -> str:
        return cast(str, self._get_key("model"))

    def maximum_context(self) -> Optional[int]:
        maximum_context = self._get_key("maximum-context")
        if maximum_context:
            return int(maximum_context)
        return None

    def file_exclude_glob_list(self) -> list[str]:
        return cast(list[str], self._get_key("file-exclude-glob-list"))

    def parser(self) -> str:
        return cast(str, self._get_key("format"))

    def _get_key(self, key: str) -> Any:
        if key in self.project_config:
            return self.project_config[key]
        elif key in self.user_config:
            return self.user_config[key]
        elif key in self.default_config:
            return self.default_config[key]
        else:
            raise ValueError(f"No value for config key {key} found")



"""
Given a codebase, a prompt and a token limit, code_context will auto-select
features to include in the code_message. This script evaluates the 
performance of that selection algorithm.

Test Guidelines:
- Codebase should be larger than the token limit
- Prompts should be thorough and specific - give it the best chance to succeed
- Tasks should draw on multiple files/features.

Scoring:
- Score based on the features (path, level, diff) auto-selected
- Use traditional precision, recall and f1 scores:
    - y_pred = [(f in code_context.features) for f in all_features]
    - y_true = [(f in test.expected) for f in all_features]

Fine-tuning:
An auto-context process might look like this:
1. Use local methods (embeddings) to select a full-context worth
    of features
2. Send the code_message of those features to gpt-3, and have it 
    return a modified features spec [(path, level, diff)] based on what's
    relevant and not. (Bonus: let gpt-3 ask for more info if it needs it)
3. Use the modified features to get_code_message for gpt-4

These tests can be used to train/score #1 or #2, but we'd expect #2 to score 
a lot higher.
"""
import os
import subprocess

import pytest

from mentat.code_context import CodeContext, CodeContextSettings
from mentat.code_file import CodeFile, CodeMessageLevel
from mentat.git_handler import GIT_ROOT, get_non_gitignored_files
from mentat.llm_api import setup_api_key

from .utils import clone_repo

pytestmark = pytest.mark.benchmark

tests = [
    {
        "name": "Mentat: replace subprocess/git with GitPython",
        "codebase_url": "http://github.com/AbanteAI/mentat",
        "codebase_name": "mentat",
        "commit": "a9f055e",
        "prompt": (
            "I want to update all the files in git_handler to use the 'Repo' class from"
            " GitPython instead of calling subprocess. Update each function in"
            " mentat/git_handler.py that calls subprocess to use Repo instead. If git"
            " is run with subprocess anywhere in the code, update those as well."
        ),
        "expected_features": [
            "mentat/git_handler.py",
            "mentat/diff_context.py:173-179",
            "tests/benchmark_test.py:118-139",
            "tests/commands_test.py:26-34",
            "tests/conftest.py:259-266",
            "tests/diff_context_test.py",
            "tests/git_handler_test.py:23-30",
            "tests/clients/terminal_client_test.py:62-93",
        ],
        "expected_edits": [
            # for other benchmarks
        ],
    },
]


@pytest.mark.asyncio
async def test_code_context_performance(
    mock_stream, mock_config, mock_code_file_manager, mock_parser
):
    setup_api_key()

    for test in tests:
        print(f"\n\n{test['codebase_name']}/ \t '{test['prompt'][:50]}...'")

        code_dir = clone_repo(test["codebase_url"], test["codebase_name"])
        os.chdir(code_dir)
        if test["commit"]:
            subprocess.run(["git", "checkout", test["commit"]])

        GIT_ROOT.set(code_dir)

        # Create a context and run get_code_message to set the features
        settings = CodeContextSettings(use_embedding=True)
        code_context = await CodeContext.create(["mentat/__init__.py"], [], settings)
        _ = await code_context.get_code_message(test["prompt"], "gpt-4", 7000)

        # Calculate y_pred and y_true
        actual = {
            f.path for f in code_context.features if f.level == CodeMessageLevel.CODE
        }
        expected_features = {
            CodeFile(f).path for f in test["expected_features"]
        }  # Ignore line numbers for now
        y_pred = [f in actual for f in get_non_gitignored_files(code_dir)]
        y_true = [f in expected_features for f in get_non_gitignored_files(code_dir)]

        _TP = sum([1 for p, t in zip(y_pred, y_true) if p and t])
        _TN = sum([1 for p, t in zip(y_pred, y_true) if not p and not t])
        _FP = sum([1 for p, t in zip(y_pred, y_true) if p and not t])
        _FN = sum([1 for p, t in zip(y_pred, y_true) if not p and t])
        print(f"True Positive:\t{_TP:.3f}")
        print(f"True Negative:\t{_TN:.3f}")
        print(f"False Positive:\t{_FP:.3f}")
        print(f"False Negative:\t{_FN:.3f}")

        precision, recall = None, None
        if (_TP + _FP) > 0:
            precision = _TP / (_TP + _FP)
            print(
                f"Precision:\t{precision:.3f}\t| How many selected features are"
                " relevant?"
            )
        if (_TP + _FN) > 0:
            recall = _TP / (_TP + _FN)
            print(
                f"Recall:\t\t{recall:.3f}\t| How many relevant features are selected?"
            )
        if precision and recall:
            f1 = 2 * precision * recall / (precision + recall)
            print(f"F1:\t\t{f1:.3f}\t| Weighted average of precision and recall")



import glob
import logging
import os
from pathlib import Path
from typing import Any, Dict

from mentat.code_file import CodeFile, parse_intervals
from mentat.config_manager import CONFIG_MANAGER
from mentat.git_handler import GIT_ROOT, get_non_gitignored_files
from mentat.session_stream import SESSION_STREAM


def expand_paths(paths: list[Path]) -> list[Path]:
    """Expand paths/globs into a list of absolute paths."""
    globbed_paths = set[str]()
    invalid_paths = set[str]()
    for path in paths:
        new_paths = glob.glob(pathname=str(path), recursive=True)
        if new_paths:
            globbed_paths.update(new_paths)
        else:
            split = str(path).rsplit(":", 1)
            p = Path(split[0])
            if len(split) > 1:
                intervals = parse_intervals(split[1])
            else:
                intervals = None
            if Path(p).exists() and intervals:
                globbed_paths.add(str(path))
            else:
                invalid_paths.add(str(path))
    return [Path(path).resolve() for path in globbed_paths]


def is_file_text_encoded(file_path: Path):
    """Checks if a file is text encoded."""
    try:
        # The ultimate filetype test
        with open(file_path, "r") as f:
            f.read()
        return True
    except UnicodeDecodeError:
        return False


def abs_files_from_list(paths: list[Path], check_for_text: bool = True):
    """Returns a set of CodeFiles from a list of paths."""
    files_direct = set[CodeFile]()
    file_paths_from_dirs = set[Path]()
    invalid_paths = list[str]()
    for path in paths:
        file = CodeFile(os.path.realpath(path))
        path = Path(file.path)
        if path.is_file():
            if check_for_text and not is_file_text_encoded(path):
                logging.info(f"File path {path} is not text encoded.")
                invalid_paths.append(str(path))
            else:
                files_direct.add(file)
        elif path.is_dir():
            nonignored_files = set(
                map(
                    lambda f: Path(os.path.realpath(path / f)),
                    get_non_gitignored_files(path),
                )
            )

            file_paths_from_dirs.update(
                filter(
                    lambda f: (not check_for_text) or is_file_text_encoded(f),
                    nonignored_files,
                )
            )

    files_from_dirs = set(CodeFile(path.resolve()) for path in file_paths_from_dirs)
    return files_direct, files_from_dirs, invalid_paths


def get_include_files(
    paths: list[Path], exclude_paths: list[Path]
) -> tuple[Dict[Path, CodeFile], list[str]]:
    """Returns a complete list of text files in a given set of include/exclude Paths."""
    git_root = GIT_ROOT.get()
    config = CONFIG_MANAGER.get()

    paths = expand_paths(paths)
    exclude_paths = expand_paths(exclude_paths)

    excluded_files_direct, excluded_files_from_dirs, _ = abs_files_from_list(
        exclude_paths, check_for_text=False
    )
    excluded_files = set(
        map(lambda f: f.path, excluded_files_direct | excluded_files_from_dirs)
    )

    glob_excluded_files = set(
        Path(os.path.join(git_root, file))
        for glob_path in config.file_exclude_glob_list()
        # If the user puts a / at the beginning, it will try to look in root directory
        for file in glob.glob(
            pathname=glob_path,
            root_dir=git_root,
            recursive=True,
        )
    )
    files_direct, files_from_dirs, invalid_paths = abs_files_from_list(
        paths, check_for_text=True
    )

    # config glob excluded files only apply to files added from directories
    files_from_dirs = [
        file
        for file in files_from_dirs
        if file.path.resolve() not in glob_excluded_files
    ]
    files_direct.update(files_from_dirs)

    files = dict[Path, CodeFile]()
    for file in files_direct:
        if file.path not in excluded_files:
            files[file.path.resolve()] = file

    return files, invalid_paths


def build_path_tree(files: list[CodeFile], git_root: Path):
    """Builds a tree of paths from a list of CodeFiles."""
    tree = dict[str, Any]()
    for file in files:
        path = os.path.relpath(file.path, git_root)
        parts = Path(path).parts
        current_level = tree
        for part in parts:
            if part not in current_level:
                current_level[part] = {}
            current_level = current_level[part]
    return tree


async def print_path_tree(
    tree: dict[str, Any], changed_files: set[Path], cur_path: Path, prefix: str = ""
):
    """Prints a tree of paths, with changed files highlighted."""
    stream = SESSION_STREAM.get()
    keys = list(tree.keys())
    for i, key in enumerate(sorted(keys)):
        if i < len(keys) - 1:
            new_prefix = prefix + "│   "
            await stream.send(f"{prefix}├── ", end="")
        else:
            new_prefix = prefix + "    "
            await stream.send(f"{prefix}└── ", end="")

        cur = cur_path / key
        star = "* " if cur in changed_files else ""
        if tree[key]:
            color = "blue"
        elif star:
            color = "green"
        else:
            color = None
        await stream.send(f"{star}{key}", color=color)
        if tree[key]:
            await print_path_tree(tree[key], changed_files, cur, new_prefix)



import gzip
import json
import os
from pathlib import Path

import numpy as np

from .code_file import CodeFile, count_feature_tokens
from .config_manager import mentat_dir_path
from .llm_api import call_embedding_api, count_tokens
from .session_stream import SESSION_STREAM
from .utils import sha256

EMBEDDING_MODEL = "text-embedding-ada-002"
EMBEDDING_MAX_TOKENS = 8192


class EmbeddingsDatabase:
    # { sha256 : [ 1536 floats ] }
    _dict: dict[str, list[float]] = dict[str, list[float]]()

    def __init__(self, output_dir: Path | None = None):
        if output_dir is None:
            output_dir = mentat_dir_path
        os.makedirs(output_dir, exist_ok=True)
        self.path = Path(output_dir) / "embeddings.json.gz"
        if self.path.exists():
            with gzip.open(self.path, "rt") as f:
                self._dict = json.load(f)

    def save(self):
        with gzip.open(self.path, "wt") as f:
            json.dump(self._dict, f)

    def __getitem__(self, key: str) -> list[float]:
        return self._dict[key]

    def __setitem__(self, key: str, value: list[float]):
        self._dict[key] = value

    def __contains__(self, key: str) -> bool:
        return key in self._dict


database = EmbeddingsDatabase()


def _batch_ffd(data: dict[str, int], batch_size: int) -> list[list[str]]:
    """Batch files using the First Fit Decreasing algorithm."""
    # Sort the data by the length of the strings in descending order
    sorted_data = sorted(data.items(), key=lambda x: x[1], reverse=True)
    batches = list[list[str]]()
    for key, value in sorted_data:
        # Place each item in the first batch that it fits in
        placed = False
        for batch in batches:
            if sum(data[k] for k in batch) + value <= batch_size:
                batch.append(key)
                placed = True
                break
        if not placed:
            batches.append([key])
    return batches


def _cosine_similarity(v1: list[float], v2: list[float]) -> float:
    """Calculate the cosine similarity between two vectors."""
    dot_product = np.dot(v1, v2)
    norm_v1 = np.linalg.norm(v1)
    norm_v2 = np.linalg.norm(v2)
    return dot_product / (norm_v1 * norm_v2)


async def get_feature_similarity_scores(
    prompt: str, features: list[CodeFile]
) -> list[float]:
    """Return the similarity scores for a given prompt and list of features."""
    global database
    stream = SESSION_STREAM.get()

    # Keep things in the same order
    checksums: list[str] = [f.get_checksum() for f in features]
    tokens: list[int] = await count_feature_tokens(features, EMBEDDING_MODEL)

    # Make a checksum:content dict of all items that need to be embedded
    items_to_embed = dict[str, str]()
    items_to_embed_tokens = dict[str, int]()
    prompt_checksum = sha256(prompt)
    if prompt_checksum not in database:
        items_to_embed[prompt_checksum] = prompt
        items_to_embed_tokens[prompt_checksum] = count_tokens(prompt, EMBEDDING_MODEL)
    for feature, checksum, token in zip(features, checksums, tokens):
        if token > EMBEDDING_MAX_TOKENS:
            continue
        if checksum not in database:
            feature_content = await feature.get_code_message()
            # Remove line numbering
            items_to_embed[checksum] = "\n".join(feature_content)
            items_to_embed_tokens[checksum] = token

    # Fetch embeddings in batches
    batches = _batch_ffd(items_to_embed_tokens, EMBEDDING_MAX_TOKENS)
    for i, batch in enumerate(batches):
        batch_content = [items_to_embed[k] for k in batch]
        await stream.send(f"Embedding batch {i}/{len(batches)}...")
        response = call_embedding_api(batch_content, EMBEDDING_MODEL)
        for k, v in zip(batch, response):
            database[k] = v
    if len(batches) > 0:
        database.save()

    # Calculate similarity score for each feature
    prompt_embedding = database[prompt_checksum]
    scores = [0.0 for _ in checksums]
    for i, checksum in enumerate(checksums):
        if checksum not in database:
            continue
        feature_embedding = database[checksum]
        scores[i] = _cosine_similarity(prompt_embedding, feature_embedding)

    return scores



import os
from pathlib import Path

from git import Repo

CLONE_TO_DIR = Path(__file__).parent / "repos"


def clone_repo(url: str, local_dir_name: str, refresh: bool = False) -> None:
    local_dir = CLONE_TO_DIR / local_dir_name
    if os.path.exists(local_dir):
        if refresh:
            repo = Repo(local_dir)
            repo.git.reset("--hard")
            repo.git.clean("-fd")
            repo.remotes.origin.pull()
    else:
        repo = Repo.clone_from(url, local_dir)
    return local_dir



from __future__ import annotations

import webbrowser
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List

from mentat.code_file_manager import CODE_FILE_MANAGER
from mentat.conversation import CONVERSATION, MessageRole
from mentat.session_stream import SESSION_STREAM
from mentat.utils import create_viewer

from .code_context import CODE_CONTEXT
from .errors import MentatError
from .git_handler import commit


class Command(ABC):
    # Unfortunately, Command isn't defined here yet, so even with annotations we need quotation marks
    _registered_commands = dict[str, type["Command"]]()
    hidden = False

    def __init_subclass__(cls, command_name: str | None) -> None:
        if command_name is not None:
            Command._registered_commands[command_name] = cls

    @classmethod
    def create_command(cls, command_name: str) -> Command:
        if command_name not in cls._registered_commands:
            return InvalidCommand(command_name)

        command_cls = cls._registered_commands[command_name]
        return command_cls()

    @classmethod
    def get_command_names(cls) -> list[str]:
        return [
            name
            for name, command in cls._registered_commands.items()
            if not command.hidden
        ]

    @classmethod
    def get_command_completions(cls) -> List[str]:
        return list(map(lambda name: "/" + name, cls.get_command_names()))

    @abstractmethod
    async def apply(self, *args: str) -> None:
        pass

    # TODO: make more robust way to specify arguments for commands
    @classmethod
    @abstractmethod
    def argument_names(cls) -> list[str]:
        pass

    @classmethod
    @abstractmethod
    def help_message(cls) -> str:
        pass


class InvalidCommand(Command, command_name=None):
    def __init__(self, invalid_name: str):
        self.invalid_name = invalid_name

    async def apply(self, *args: str) -> None:
        await SESSION_STREAM.get().send(
            f"{self.invalid_name} is not a valid command. Use /help to see a list of"
            " all valid commands",
            color="light_yellow",
        )

    @classmethod
    def argument_names(cls) -> list[str]:
        raise MentatError("Argument names called on invalid command")

    @classmethod
    def help_message(cls) -> str:
        raise MentatError("Help message called on invalid command")


help_message_width = 60


class HelpCommand(Command, command_name="help"):
    async def apply(self, *args: str) -> None:
        stream = SESSION_STREAM.get()

        if not args:
            commands = Command.get_command_names()
        else:
            commands = args
        for command_name in commands:
            if command_name not in Command._registered_commands:
                await stream.send(
                    f"Error: Command {command_name} does not exist.", color="red"
                )
            else:
                command_class = Command._registered_commands[command_name]
                argument_names = command_class.argument_names()
                help_message = command_class.help_message()
                message = (
                    " ".join(
                        [f"/{command_name}"] + [f"<{arg}>" for arg in argument_names]
                    ).ljust(help_message_width)
                    + help_message
                )
                await stream.send(message)

    @classmethod
    def argument_names(cls) -> list[str]:
        return []

    @classmethod
    def help_message(cls) -> str:
        return "Displays this message"


class CommitCommand(Command, command_name="commit"):
    default_message = "Automatic commit"

    async def apply(self, *args: str) -> None:
        if args:
            commit(args[0])
        else:
            commit(self.__class__.default_message)

    @classmethod
    def argument_names(cls) -> list[str]:
        return [f"commit_message={cls.default_message}"]

    @classmethod
    def help_message(cls) -> str:
        return "Commits all of your unstaged and staged changes to git"


class IncludeCommand(Command, command_name="include"):
    async def apply(self, *args: str) -> None:
        stream = SESSION_STREAM.get()
        code_context = CODE_CONTEXT.get()

        if len(args) == 0:
            await stream.send("No files specified\n", color="yellow")
            return
        for file_path in args:
            invalid_paths = code_context.include_file(Path(file_path).absolute())
            for invalid_path in invalid_paths:
                await stream.send(
                    f"File path {invalid_path} is not text encoded, and was skipped.",
                    color="light_yellow",
                )
            if file_path not in invalid_paths:
                await stream.send(f"{file_path} added to context", color="green")

    @classmethod
    def argument_names(cls) -> list[str]:
        return ["file1", "file2", "..."]

    @classmethod
    def help_message(cls) -> str:
        return "Add files to the code context"


class ExcludeCommand(Command, command_name="exclude"):
    async def apply(self, *args: str) -> None:
        stream = SESSION_STREAM.get()
        code_context = CODE_CONTEXT.get()

        if len(args) == 0:
            await stream.send("No files specified\n", color="yellow")
            return
        for file_path in args:
            code_context.exclude_file(Path(file_path).absolute())
            await stream.send(f"{file_path} removed from context", color="green")

    @classmethod
    def argument_names(cls) -> list[str]:
        return ["file1", "file2", "..."]

    @classmethod
    def help_message(cls) -> str:
        return "Remove files from the code context"


class UndoCommand(Command, command_name="undo"):
    async def apply(self, *args: str) -> None:
        stream = SESSION_STREAM.get()
        code_file_manager = CODE_FILE_MANAGER.get()
        errors = code_file_manager.history.undo()
        if errors:
            await stream.send(errors)
        await stream.send("Undo complete", color="green")

    @classmethod
    def argument_names(cls) -> list[str]:
        return []

    @classmethod
    def help_message(cls) -> str:
        return "Undo the last change made by Mentat"


class UndoAllCommand(Command, command_name="undo-all"):
    async def apply(self, *args: str) -> None:
        stream = SESSION_STREAM.get()
        code_file_manager = CODE_FILE_MANAGER.get()
        errors = code_file_manager.history.undo_all()
        if errors:
            await stream.send(errors)
        await stream.send("Undos complete", color="green")

    @classmethod
    def argument_names(cls) -> list[str]:
        return []

    @classmethod
    def help_message(cls) -> str:
        return "Undo all changes made by Mentat"


class ClearCommand(Command, command_name="clear"):
    async def apply(self, *args: str) -> None:
        stream = SESSION_STREAM.get()
        conversation = CONVERSATION.get()
        # Only keep system messages (for now just the prompt)
        conversation.messages = [
            message
            for message in conversation.messages
            if message["role"] == MessageRole.System
        ]
        message = "Message history cleared"
        await stream.send(message, color="green")

    @classmethod
    def argument_names(cls) -> list[str]:
        return []

    @classmethod
    def help_message(cls) -> str:
        return "Clear the current conversation's message history"


class ConversationCommand(Command, command_name="conversation"):
    hidden = True

    async def apply(self, *args: str) -> None:
        conversation = CONVERSATION.get()
        viewer_path = create_viewer(conversation.literal_messages)
        webbrowser.open(f"file://{viewer_path.resolve()}")

    @classmethod
    def argument_names(cls) -> list[str]:
        return []

    @classmethod
    def help_message(cls) -> str:
        return "Opens an html page showing the conversation as seen by Mentat so far"
